/**
 * LICENCIA LGPL:
 * 
 * Esta librera es Software Libre; Usted puede redistribuirla y/o modificarla
 * bajo los trminos de la GNU Lesser General Public License (LGPL) tal y como 
 * ha sido publicada por la Free Software Foundation; o bien la versin 2.1 de 
 * la Licencia, o (a su eleccin) cualquier versin posterior.
 * 
 * Esta librera se distribuye con la esperanza de que sea til, pero SIN 
 * NINGUNA GARANTA; tampoco las implcitas garantas de MERCANTILIDAD o 
 * ADECUACIN A UN PROPSITO PARTICULAR. Consulte la GNU Lesser General Public 
 * License (LGPL) para ms detalles
 * 
 * Usted debe recibir una copia de la GNU Lesser General Public License (LGPL) 
 * junto con esta librera; si no es as, escriba a la Free Software Foundation 
 * Inc. 51 Franklin Street, 5 Piso, Boston, MA 02110-1301, USA o consulte
 * <http://www.gnu.org/licenses/>.
 *
 * Copyright 2011 Agencia de Tecnologa y Certificacin Electrnica
 */
package es.accv.arangi.base.request;

import java.io.IOException;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.X509EncodedKeySpec;

import org.apache.log4j.Logger;
import org.bouncycastle.asn1.ASN1Encoding;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;

import sun.security.rsa.RSAPublicKeyImpl;

import com.novosec.pkix.asn1.crmf.CertRequest;

import es.accv.arangi.base.exception.CertificateRequestException;
import es.accv.arangi.base.util.Util;

/**
 * Representacin de una peticin de certificado en formato Certification Request
 * Message Format (CRMF) segn la <a href="http://www.ietf.org/rfc/rfc4211.txt">
 * RFC-4211</a>.
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos Manuel Gutirrez Nez</a>
 *
 */
public class CRMFRequest extends CertificateRequest {
	
	/*
	 * Logger de la clase
	 */
	Logger logger = Logger.getLogger(CRMFRequest.class);
	
	CertRequest crmfRequest;
	
	byte[] crmfRequestBytes; 
	
	//-- Constructor
	
	/**
	 * Constructor al que se le pasa el fichero CRMF en base64
	 * 
	 * @param crmfBase64 CRMF en base64
	 * @throws IOException Excepcin si el CRMF no tiene la estructura
	 * 	correcta
	 */
	public CRMFRequest (String crmfBase64) throws IOException {
		super();
		
	    logger.debug("[PKCS10Request]::Entrada::" + crmfBase64);
	    
	    if ( crmfBase64 == null || crmfBase64.equals("") ) {
	      throw new IOException("El CRMF est vaco");
	    }
	    
	    try {
		    this.crmfRequestBytes = Util.decodeBase64(crmfBase64);
		    
		    ASN1InputStream in = new ASN1InputStream(this.crmfRequestBytes);
		    ASN1Sequence crmfSeq = (ASN1Sequence) in.readObject();
		    ASN1Sequence reqSeq = (ASN1Sequence) ((ASN1Sequence) crmfSeq.getObjectAt(0)).getObjectAt(0);
		    crmfRequest = new CertRequest(reqSeq);
		    logger.debug("[PKCS10Request]::Obtenido el CRMF"); 
	    } catch (Exception e) {
	    	throw new IOException("No se puede parsear el CRMF: " + e.getMessage());
	    }

	}
	
	//-- Implementacin de la clase abstracta
	
	@Override
	public PublicKey getPublicKey() throws CertificateRequestException {
		SubjectPublicKeyInfo keyInfo = crmfRequest.getCertTemplate().getPublicKey();
		try {
			return KeyFactory.getInstance("RSA").generatePublic(new X509EncodedKeySpec(keyInfo.getEncoded(ASN1Encoding.DER)));
		} catch (InvalidKeySpecException e) {
			logger.info("[CRMFRequest.getPublicKey]::Error por especificaciones de la clave pblica no vlidas");
			throw new CertificateRequestException ("Error por especificaciones de la clave pblica no vlidas", e);
		} catch (NoSuchAlgorithmException e) {
			logger.info("[CRMFRequest.getPublicKey]::No existe el algoritmo de la clave pblica");
			throw new CertificateRequestException ("No existe el algoritmo de la clave pblica", e);
		} catch (IOException e) {
			logger.info("[CRMFRequest.getPublicKey]::No ha sido posible obtener el array de bytes de la clave pblica");
			throw new CertificateRequestException ("No ha sido posible obtener el array de bytes de la clave pblica", e);
		}
	}

	@Override
	public int getPublicKeyLength() throws CertificateRequestException {
		return ((RSAPublicKeyImpl)getPublicKey()).getModulus().bitLength();
	}
	
	@Override
	public byte[] getEncoded() {
		return this.crmfRequestBytes;
	}


}
