/**
 * LICENCIA LGPL:
 * 
 * Esta librera es Software Libre; Usted puede redistribuirla y/o modificarla
 * bajo los trminos de la GNU Lesser General Public License (LGPL) tal y como 
 * ha sido publicada por la Free Software Foundation; o bien la versin 2.1 de 
 * la Licencia, o (a su eleccin) cualquier versin posterior.
 * 
 * Esta librera se distribuye con la esperanza de que sea til, pero SIN 
 * NINGUNA GARANTA; tampoco las implcitas garantas de MERCANTILIDAD o 
 * ADECUACIN A UN PROPSITO PARTICULAR. Consulte la GNU Lesser General Public 
 * License (LGPL) para ms detalles
 * 
 * Usted debe recibir una copia de la GNU Lesser General Public License (LGPL) 
 * junto con esta librera; si no es as, escriba a la Free Software Foundation 
 * Inc. 51 Franklin Street, 5 Piso, Boston, MA 02110-1301, USA o consulte
 * <http://www.gnu.org/licenses/>.
 *
 * Copyright 2011 Agencia de Tecnologa y Certificacin Electrnica
 */
package es.accv.arangi.base.signature;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.util.List;

import es.accv.arangi.base.certificate.Certificate;
import es.accv.arangi.base.certificate.validation.CAList;
import es.accv.arangi.base.certificate.validation.CertificateValidationService;
import es.accv.arangi.base.util.validation.ValidationResult;
import es.accv.arangi.base.document.IDocument;
import es.accv.arangi.base.exception.certificate.NormalizeCertificateException;
import es.accv.arangi.base.exception.document.HashingException;
import es.accv.arangi.base.exception.signature.NoDocumentToSignException;
import es.accv.arangi.base.exception.signature.SignatureException;

/**
 * Interfaz que deben cumplir todos las clases que representan firmas digitales.<br><br>
 * 
 * En la medida de lo posible se ha intentado distinguir entre los bytes de la
 * firma, resultantes de los procesos de hashing ms encriptado, de las firmas
 * digitales donde se guardan, adems de estos bytes de firma, los certificados, 
 * sellos de tiempo y cualquier otra informacin que sirva a su validacin. Las 
 * clases que implementan esta interfaz representan a los segundos. Los mtodos 
 * de firma de las clases que extienden {@link es.accv.arangi.base.device.DeviceManager DeviceManager} 
 * devuelven slo los bytes de la firma.  
 * 
 * @author <a href="mailto:jgutierrez@accv.es">Jos M Gutirrez</a>
 */
public interface ISignature {

	/**
	 * Devuelve los certificados con los que se ha realizado la firma
	 * 
	 * @return Certificados con los que se ha realizado la firma
	 */
	public Certificate[] getCertificates ();
	
	/**
	 * Comprueba que las firmas son correctas en firmas attached, sin validar los certificados 
	 * de las mismas.<br><br>
	 * 
	 * IMPORTANTE: este mtodo slo puede ser utilizado si la firma es attached (el documento
	 * que origin la firma se incluye en sta). Si no es as utilizar este mismo mtodo pero 
	 * pasndole el documento que origin la firma.
	 * 
	 * @return Para cada certificado resultado de validar la firma
	 * @throws SignatureException Error tratando el objeto firma
	 * @throws HashingException Error obteniendo el hash del documento
	 * @throws NoDocumentToSignException La firma no es attached por lo que no hay documento con
	 * 	el que validarla. Utilizar este mismo mtodo pero pasndole el documento que origin la
	 * 	firma
	 */
	public ValidationResult[] isValidSignatureOnly () throws HashingException, SignatureException, NoDocumentToSignException;
	
	/**
	 * Comprueba que las firmas son correctas en firmas attached y sus certificados son vlidos. <br><br>
	 * 
	 * Sobre la validacin de certificados hay que tener en cuenta:<br>
	 * <ul>
	 * 	<li>Si existe sello de tiempos, ste slo ser valido mientras el certificado no
	 * 	haya caducado. Despus, es imposible obtener la informacin de revocacin por lo
	 * 	que este mtodo devolver un resultado falso</li>
	 * 	<li>Si la firma incluye informacin de revocacin (CRLs o respuestas OCSP) sta 
	 * 	si que se tendr en cuenta aunque el certificado haya caducado: concepto de
	 * 	firma longeva</li>
	 * </ul><br><br>
	 * 
	 * IMPORTANTE: este mtodo slo puede ser utilizado si la firma es attached (el documento
	 * que origin la firma se incluye en sta). Si no es as utilizar este mismo mtodo pero 
	 * pasndole el documento que origin la firma.
	 * 
	 * @param caList Lista de certificados de CA admitidos por la aplicacin que usa
	 * 	Arangi
	 * @return Para cada certificado resultado de comprobar si la firma es correcta y el certificado es
	 * 	vlido
	 * @throws SignatureException Error tratando el objeto firma
	 * @throws HashingException Error obteniendo el hash del documento
	 * @throws NormalizeCertificateException Alguno de los certificados no puede ser 
	 * 	normalizado al formato reconocido por el proveedor criptogrfico de Arang o su 
	 * 	firma no es correcta o no puede ser analizada
	 * @throws NoDocumentToSignException La firma no es attached por lo que no hay documento con
	 * 	el que validarla. Utilizar este mismo mtodo pero pasndole el documento que origin la
	 * 	firma
	 */
	public ValidationResult[] isValid (CAList caList) throws HashingException, SignatureException, NormalizeCertificateException, NoDocumentToSignException;
	
	/**
	 * Comprueba que las firmas son correctas en firmas attached y sus certificados son vlidos. <br><br>
	 * 
	 * Sobre la validacin de certificados hay que tener en cuenta:<br>
	 * <ul>
	 * 	<li>Si existe sello de tiempos, ste slo ser valido mientras el certificado no
	 * 	haya caducado. Despus, es imposible obtener la informacin de revocacin por lo
	 * 	que este mtodo devolver un resultado falso</li>
	 * 	<li>Si la firma incluye informacin de revocacin (CRLs o respuestas OCSP) sta 
	 * 	si que se tendr en cuenta aunque el certificado haya caducado: concepto de
	 * 	firma longeva</li>
	 * <li>La validacin se llevar a cabo a travs de los servicios de validacin
	 * 	pasados como parmetro.</li>
	 * </ul><br><br>
	 * 
	 * IMPORTANTE: este mtodo slo puede ser utilizado si la firma es attached (el documento
	 * que origin la firma se incluye en sta). Si no es as utilizar este mismo mtodo pero 
	 * pasndole el documento que origin la firma.
	 * 
	 * @param validationServices Lista de servicios de validacin
	 * @return Para cada certificado resultado de comprobar si la firma es correcta y el certificado es
	 * 	vlido
	 * @throws SignatureException Error tratando el objeto firma
	 * @throws HashingException Error obteniendo el hash del documento
	 * @throws NormalizeCertificateException Alguno de los certificados no puede ser 
	 * 	normalizado al formato reconocido por el proveedor criptogrfico de Arang o su 
	 * 	firma no es correcta o no puede ser analizada
	 * @throws NoDocumentToSignException La firma no es attached por lo que no hay documento con
	 * 	el que validarla. Utilizar este mismo mtodo pero pasndole el documento que origin la
	 * 	firma
	 */
	public ValidationResult[] isValid (List<CertificateValidationService> validationServices) throws HashingException, SignatureException, NormalizeCertificateException, NoDocumentToSignException;
	
	/**
	 * Comprueba que las firmas son correctas, sin validar los certificados de las mismas.
	 * 
	 * @param document Documento que origin la firma
	 * @return Para cada certificado resultado de validar la firma
	 * @throws SignatureException Error tratando el objeto firma
	 * @throws HashingException Error obteniendo el hash del documento
	 */
	public ValidationResult[] isValidSignatureOnly (IDocument document) throws HashingException, SignatureException;
	
	/**
	 * Comprueba que las firmas son correctas y sus certificados son vlidos.<br><br> 
	 * 
	 * Sobre la validacin de certificados hay que tener en cuenta:<br>
	 * <ul>
	 * 	<li>Si existe sello de tiempos, ste slo ser valido mientras el certificado no
	 * 	haya caducado. Despus, es imposible obtener la informacin de revocacin por lo
	 * 	que este mtodo devolver un resultado falso</li>
	 * 	<li>Si la firma incluye informacin de revocacin (CRLs o respuestas OCSP) sta 
	 * 	si que se tendr en cuenta aunque el certificado haya caducado: concepto de
	 * 	firma longeva</li>
	 * </ul>
	 * 
	 * @param document Documento que origin la firma
	 * @param caList Lista de certificados de CA admitidos por la aplicacin que usa
	 * 	Arangi
	 * @return Para cada certificado resultado de comprobar si la firma es correcta y el certificado es
	 * 	vlido
	 * @throws SignatureException Error tratando el objeto firma
	 * @throws HashingException Error obteniendo el hash del documento
	 * @throws NormalizeCertificateException Alguno de los certificados no puede ser 
	 * 	normalizado al formato reconocido por el proveedor criptogrfico de Arangi o su 
	 * 	firma no es correcta o no puede ser analizada
	 */
	public ValidationResult[] isValid (IDocument document, CAList caList) throws HashingException, SignatureException, NormalizeCertificateException;
	
	/**
	 * Comprueba que las firmas son correctas y sus certificados son vlidos.<br><br> 
	 * 
	 * Sobre la validacin de certificados hay que tener en cuenta:<br>
	 * <ul>
	 * 	<li>Si existe sello de tiempos, ste slo ser valido mientras el certificado no
	 * 	haya caducado. Despus, es imposible obtener la informacin de revocacin por lo
	 * 	que este mtodo devolver un resultado falso</li>
	 * 	<li>Si la firma incluye informacin de revocacin (CRLs o respuestas OCSP) sta 
	 * 	si que se tendr en cuenta aunque el certificado haya caducado: concepto de
	 * 	firma longeva</li>
	 * <li>La validacin se llevar a cabo a travs de los servicios de validacin
	 * 	pasados como parmetro.</li>
	 * </ul>
	 * 
	 * @param document Documento que origin la firma
	 * @param validationServices Lista de servicios de validacin
	 * @return Para cada certificado resultado de comprobar si la firma es correcta y el certificado es
	 * 	vlido
	 * @throws SignatureException Error tratando el objeto firma
	 * @throws HashingException Error obteniendo el hash del documento
	 * @throws NormalizeCertificateException Alguno de los certificados no puede ser 
	 * 	normalizado al formato reconocido por el proveedor criptogrfico de Arangi o su 
	 * 	firma no es correcta o no puede ser analizada
	 */
	public ValidationResult[] isValid (IDocument document, List<CertificateValidationService> validationServices) throws HashingException, SignatureException, NormalizeCertificateException;
	
	/**
	 * Guarda la firma en disco
	 * 
	 * @param file Fichero donde se guardar la firma
	 * @throws IOException Errores de entrada / salida
	 */
	public void save (File file) throws IOException;
	
	/**
	 * Guarda la firma en un stream de escritura.
	 * 
	 * @param out Stream de escritura
	 * @throws IOException Errores de entrada / salida
	 */
	public void save (OutputStream out) throws IOException;
	
	/**
	 * Devuelve una cadena de texto con el tipo de la firma
	 * 
	 * @return Cadena de texto con el tipo de la firma
	 */
	public String getSignatureType ();
}
